/**
 * 
 */
package de.pdark.decentxml;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import de.pdark.decentxml.XMLTokenizer.Type;

public class Element implements Node
{
    private XMLTokenizer.Token startToken;
    private String beginName;
    private String endName;
    private String name;
    private List<Node> attributes;
    private List<Node> children;
    private boolean compactEmpty;
    
    public Element (XMLTokenizer.Token token)
    {
        this.startToken = token;
        this.beginName = token.getText ();
        this.beginName = this.beginName.substring (1);
        this.name = this.beginName.trim ();
    }
    
    public Element (String name)
    {
        this.name = name;
        compactEmpty = true;
    }
    
    public XMLTokenizer.Token getStartToken ()
    {
        return startToken;
    }

    public int getStartOffset ()
    {
        return startToken == null ? -1 : startToken.getStartOffset ();
    }
    
    public int getEndOffset ()
    {
        return startToken == null ? -1 : startToken.getEndOffset ();
    }
    
    public String getBeginName ()
    {
        return beginName == null ? name : beginName;
    }
    
    public void setBeginName (String beginName)
    {
        this.beginName = beginName;
    }
    
    public String getEndName ()
    {
        return endName == null ? name : endName;
    }
    
    public void setEndName (String endName)
    {
        this.endName = endName;
    }
    
    public String getName ()
    {
        return name;
    }
    
    public void addAttribute (Node node)
    {
        // Only Text and Attribute allowed.
        
        if (attributes == null)
            attributes = new ArrayList<Node> ();
        attributes.add (node);
    }
    
    public List<Attribute> getAttributes ()
    {
        if (attributes == null)
            return Collections.emptyList ();
        
        List<Attribute> result = new ArrayList<Attribute> ();
        
        for (Node n: attributes)
        {
            if (n instanceof Attribute)
                result.add ((Attribute)n);
        }
        
        return result;
    }
    
    public Map<String, Attribute> getAttributeMap ()
    {
        if (attributes == null)
            return Collections.emptyMap ();
        
        Map<String, Attribute> result = new LinkedHashMap<String, Attribute> ();
        
        for (Node n: attributes)
        {
            if (n instanceof Attribute)
            {
                Attribute a = (Attribute)n;
                result.put (a.getName (), a);
            }
        }
        
        return result;
    }
    
    public Attribute getAttribute (String name)
    {
        if (attributes == null)
            return null;
        
        for (Node n: attributes)
        {
            if (n instanceof Attribute)
            {
                Attribute a = (Attribute)n;
                if (name.equals (a.getName ()))
                    return a;
            }
        }

        return null;
    }
    
    public String getAttributeValue (String name)
    {
        Attribute a = getAttribute (name);
        return a == null ? null : a.getAttributeValue ();
    }
    
    public void addChild (Node node)
    {
        addChild (children == null ? 0 : children.size (), node);
    }
    
    public void addChild (int index, Node node)
    {
        if (node instanceof Attribute)
            throw new XMLParseException ("Can't add an attribute as a child of an element", startToken);
        
        if (children == null)
            children = new ArrayList<Node> ();
        children.add (index, node);
    }
    
    public void setCompactEmpty (boolean compactEmpty)
    {
        this.compactEmpty = compactEmpty;
    }
    
    public boolean isCompactEmpty ()
    {
        return compactEmpty && !hasChildren ();
    }
    
    public void toXML (Writer writer) throws IOException
    {
        writer.append ("<");
        writer.append (getBeginName ());
        
        if (attributes != null)
        {
            for (Node n: attributes)
                n.toXML (writer);
        }
        
        if (!hasChildren ())
        {
            if (compactEmpty)
                writer.append ("/>");
            else
            {
                writer.append ("></");
                writer.append (getEndName ());
                writer.append (">");
            }
            
            return;
        }
        
        writer.append (">");
        
        if (children != null)
        {
            for (Node n: children)
                n.toXML (writer);
        }

        writer.append ("</");
        writer.append (getEndName ());
        writer.append (">");
    }

    public String toXML ()
    {
        return BasicNode.toXML (this);
    }
    
    @Override
    public String toString ()
    {
        return toXML ();
    }
    
    public Type getType ()
    {
        return Type.ELEMENT;
    }
    
    public Element getChild (String name)
    {
        if (children == null)
            return null;
        
        for (Node n: children)
        {
            if (n instanceof Element)
            {
                Element e = (Element)n;
                if (name.equals (e.getName ()))
                    return e;
            }
        }
        
        return null;
    }
    
    public boolean hasChildren ()
    {
        return children != null && children.size () != 0;
    }
    
    public List<Node> getNodes ()
    {
        if (children == null)
            return Collections.emptyList ();

        return children;
    }
    
    public Node getNode (int index)
    {
        return children.get (index);
    }
    
    public Node removeNode (int index)
    {
        return children.remove (index);
    }
    
    public boolean removeNode (Node n)
    {
        if (children == null)
            return false;
        
        return children.remove (n);
    }
    
    public List<Node> getNodes (NodeFilter filter)
    {
        if (children == null)
            return Collections.emptyList ();
        
        List<Node> result = new ArrayList<Node> (children.size ());
        
        for (Node n: children)
        {
            if (filter.matches (n))
            {
                result.add ((Element)n);
            }
        }
        
        return result;
    }
    
    public final static NodeFilter ELEMENT_FILTER = new NodeFilter () {
        @Override
        public boolean matches (Node n)
        {
            return n instanceof Element;
        }
    };
    
    @SuppressWarnings("unchecked")
    public List<Element> getChildren ()
    {
        return (List)getNodes (ELEMENT_FILTER);
    }
    
    @SuppressWarnings("unchecked")
    public List<Element> getChildren (final String name)
    {
        NodeFilter nameFilter = new NodeFilter () {
            @Override
            public boolean matches (Node n)
            {
                return n instanceof Element
                    && name.equals (((Element)n).getName ());
            }
        };
        
        return (List)getNodes (nameFilter);
    }

    public void clearChildNodes ()
    {
        if (children != null)
            children.clear ();
    }

    public String getText ()
    {
        if (children == null)
            return "";
        
        StringBuilder buffer = new StringBuilder ();
        for (Node n: children)
        {
            if (n instanceof Text)
                buffer.append (((Text)n).getValue ());
            else if (n instanceof Element)
                buffer.append (((Element)n).getText ());
        }
        
        return buffer.toString ();
    }
    
    public String getTrimmedText ()
    {
        return getText ().trim ();
    }
    
    public String getNormalizedText ()
    {
        return getTrimmedText ().replaceAll ("\\s+", " ");
    }
    
    public void setText (String text)
    {
        clearText ();
        addChild (0, new Text (text));
    }

    public void clearText ()
    {
        if (children == null)
            return;
        
        for (Iterator<Node> iter = children.iterator (); iter.hasNext (); )
        {
            Node n = iter.next ();
            
            if (n instanceof Text)
            {
                iter.remove ();
            }
            else if (n instanceof Element)
            {
                ((Element)n).clearText ();
            }
        }
    }
}