package de.pdark.decentxml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Map;

public class XMLInputStreamReader extends Reader
{
    public final static String ENCODING_UTF_8 = "UTF-8";
    public final static String ENCODING_ISO_Latin_1 = "ISO-8859-1";
    
    public static Map<String, String> ENCODING_MAP = new HashMap<String, String> ();
    static
    {
        ENCODING_MAP.put ("Utf-8", "UTF-8");
        ENCODING_MAP.put ("utf-8", "UTF-8");
        ENCODING_MAP.put ("Latin-1", "ISO-8859-1");
    }
    
    private InputStream in;
    private String xmlEncoding;
    private String javaEncoding;
    private Reader delegate;
    
    public XMLInputStreamReader (InputStream in)
    {
        this.in = in;
    }
    
    public void determineEncoding ()
    {
        if (xmlEncoding != null)
            return;
        
        byte[] header = new byte[1024];
        int headerLength;
        try
        {
            headerLength = in.read (header);
        }
        catch (IOException e)
        {
            throw new XMLParseException ("Error parsing XML declaration: "+e.getMessage (), e);
        }
        
        String encoding = ENCODING_UTF_8;
        if (headerLength < 2)
        {
            // Do nothing
        }
        else if ((header[0] == (byte)0xFE && header[1] == (byte)0xFF)
                || (header[0] == (byte)0xFF && header[1] == (byte)0xFE))
        {
            encoding = "UTF-16";
        }
        else if (headerLength < 4)
        {
            // Do nothing
        }
        else if (header[0] == '<' && header[1] == '?' && header[2] == 'x' && header[3] == 'm')
        {
            int pos = 4;
            for ( ; pos<headerLength; pos ++)
            {
                int c = header[pos] & 0xFF;
                if (c == '>')
                    break;
                
                if (Character.isLetter (c))
                    continue;
                if (Character.isDigit (c))
                    continue;
                if (Character.isWhitespace (c))
                    continue;
                if (c == '?' || c == '"' || c == '\'' || c == '=' || c == '-' || c == '_' || c == '.')
                    continue;
                
                throw new XMLParseException ("Found illegal character in XML header: '"+(char)c+"'");
            }
            
            String s;
            try
            {
                s = new String (header, 0, pos, ENCODING_UTF_8);
            }
            catch (UnsupportedEncodingException e) // Note: this code will never be executed
            {
                throw new XMLParseException ("Error parsing XML declaration: Unsupported encoding "+e.getMessage (), e);
            }
            //System.out.println (s);

            if (!s.contains ("version=\"1.0\"")
                    && !s.contains ("version='1.0'"))
                throw new XMLParseException ("Missing version in XML declaration");

            String pattern = "encoding=";
            pos = s.indexOf (pattern);
            if (pos != -1)
            {
                pos = pos + pattern.length ();
                int c = s.charAt (pos);
                if (c != '\'' && c != '"')
                    throw new XMLParseException ("Missing opening quote for encoding");
                
                pos ++;
                int pos2 = s.indexOf (c, pos);
                if (pos2 == -1)
                    throw new XMLParseException ("Missing closing quote for encoding");
                
                encoding = s.substring (pos, pos2);
                
            }
        }
        
        //System.out.println ("encoding="+encoding);
        
        JoinedInputStream jin = new JoinedInputStream ()
        .add (new ByteArrayInputStream (header, 0, headerLength))
        .add (in)
        ;

        setXmlEncoding (encoding);
        try
        {
            delegate = new InputStreamReader (jin, javaEncoding);
        }
        catch (UnsupportedEncodingException e)
        {
            throw new XMLParseException ("Error parsing XML declaration: Unsupported encoding '"+javaEncoding+"'", e);
        }
    }

    public String getXmlEncoding ()
    {
        determineEncoding ();
        return xmlEncoding;
    }
    
    public void setXmlEncoding (String xmlEncoding)
    {
        this.xmlEncoding = xmlEncoding;
        javaEncoding = ENCODING_MAP.get (xmlEncoding);
        if (javaEncoding == null)
            javaEncoding = xmlEncoding;
    }
    
    public String getJavaEncoding ()
    {
        determineEncoding ();
        return javaEncoding;
    }
    
    @Override
    public void close () throws IOException
    {
        if (delegate == null)
            in.close ();
        else
            delegate.close ();
    }

    @Override
    public int read (char[] cbuf, int off, int len) throws IOException
    {
        determineEncoding ();
        return delegate.read (cbuf, off, len);
    }
}
