package de.pdark.decentxml;

public class XMLTokenizer
{
    public static enum Type {
        TEXT, CDATA,
        /** Whitespace inside of an element tag */
        ELEMENT_WHITESPACE,
        PROCESSING_INSTRUCTION,
        COMMENT,
        /** This is the name part of the start tag */
        BEGIN_ELEMENT,
        /** An attribute in the start tag */
        ATTRIBUTE,
        /** The token which terminates the start tag. It's value is either '>' or '/>' */
        BEGIN_ELEMENT_END,
        /** The end element */
        END_ELEMENT,
        /** A document node */
        DOCUMENT,
        /** Node-type for elements after they have been parsed */
        ELEMENT
    }
    
    private final XMLSource source;
    private int pos;
    private boolean inStartElement;
    
    public static class Token
    {
        private XMLSource source;
        private Type type;
        private int startOffset;
        private int endOffset;
        
        public void setSource (XMLSource source)
        {
            this.source = source;
        }

        public XMLSource getSource ()
        {
            return source;
        }

        public String typeToString ()
        {
            if (type == null)
                return "***";
            
            switch (getType())
            {
            case TEXT: return "TEXT";
            case PROCESSING_INSTRUCTION: return "PI";
            case COMMENT: return "COMMENT";
            case BEGIN_ELEMENT: return "BEGIN_ELEMENT";
            case BEGIN_ELEMENT_END: return "BEGIN_ELEMENT_END";
            case ATTRIBUTE: return "ATTRIBUTE";
            case END_ELEMENT: return "END_ELEMENT";
            case ELEMENT_WHITESPACE: return "ELEMENT_WHITESPACE";
            case CDATA: return "CDATA";
            }
            
            // Note: this code will never be executed unless a new token was added and the
            // switch wasn't updated.
            return "***";
        }
        
        public String getText ()
        {
            return getSource() == null ? null : getSource().substring (getStartOffset(), getEndOffset());
        }
        
        public String getEscapedText ()
        {
            String s = getText ();
            if (s == null)
                return null;
            
            StringBuilder buffer = new StringBuilder ();
            for (int i=0; i<s.length (); i++)
            {
                char c = s.charAt (i);
                switch (c)
                {
                case '\n': buffer.append ("\\n"); break;
                case '\r': buffer.append ("\\r"); break;
                case '\t': buffer.append ("\\t"); break;
                case '"': buffer.append ("\\\""); break;
                default:
                    if (c < 16)
                    {
                        buffer.append ("\\u000");
                        buffer.append (Integer.toHexString (c));
                    }
                    else if (c < 32 || (c >= 128 && c < 161))
                    {
                        buffer.append ("\\u00");
                        buffer.append (Integer.toHexString (c));
                    }
                    else if (c > 256 && c < 0x1000)
                    {
                        buffer.append ("\\u0");
                        buffer.append (Integer.toHexString (c));
                    }
                    else if (c >= 0x1000)
                    {
                        buffer.append ("\\u");
                        buffer.append (Integer.toHexString (c));
                    }
                    else
                        buffer.append (c);
                }
            }
            return buffer.toString ();
        }
        
        @Override
        public String toString ()
        {
            return "Token (" + typeToString () + ", " + getStartOffset() + ":" + getEndOffset() + ", \"" + getEscapedText () + "\")";
        }

        public void setType (Type type)
        {
            this.type = type;
        }

        public Type getType ()
        {
            return type;
        }

        public void setStartOffset (int startOffset)
        {
            this.startOffset = startOffset;
        }

        public int getStartOffset ()
        {
            return startOffset;
        }

        public void setEndOffset (int endOffset)
        {
            this.endOffset = endOffset;
        }

        public int getEndOffset ()
        {
            return endOffset;
        }
    }
    
    public XMLTokenizer (XMLSource source)
    {
        this.source = source;
    }
    
    public Token next ()
    {
        if (pos >= source.length ())
            return null;
        
        Token token = new Token ();
        token.setSource (source);
        token.setStartOffset (pos);
        
        char c = source.charAt (pos);
        if (inStartElement)
        {
            if (c == '>')
            {
                pos ++;
                token.setType (Type.BEGIN_ELEMENT_END);
                inStartElement = false;
            }
            else if (c == '/') // Empty element
            {
                pos ++;
                if (pos >= source.length () || source.charAt (pos) != '>')
                    throw new XMLParseException ("Expected '/>'", source, pos-1);
                
                pos ++;
                
                token.setType (Type.BEGIN_ELEMENT_END);
                inStartElement = false;
            }
            else if (Character.isWhitespace (c))
            {
                token.setType (Type.ELEMENT_WHITESPACE);
                while (Character.isWhitespace (source.charAt (pos)))
                {
                    pos ++;
                }
            }
            else
            {
                parseAttribute (token);
            }
        }
        else if (c == '<')
        {
            pos ++;
            parseBeginSomething (token);
        }
        else
        {
            parseText (token);
        }
        
        token.setEndOffset (pos);
        
        return token;
    }

    public XMLSource getSource ()
    {
        return source;
    }
    
    public int getOffset ()
    {
        return pos;
    }
    
    // <tag, <?pi, <!--, <![CDATA[, </tag
    private void parseBeginSomething (Token token)
    {
        char c = source.charAt (pos);
        if (c == '?')
        {
            pos ++;
            parseProcessingInstruction (token);
        }
        else if (c == '!')
        {
            pos ++;
            parseExcalamtion (token);
        }
        else if (c == '/')
        {
            pos ++;
            parseEndElement (token);
        }
        else
            parseBeginElement (token);
    }

    private void parseBeginElement (Token token)
    {
        token.setType (Type.BEGIN_ELEMENT);
        inStartElement = true;
        
        skipWhiteSpace ();
        
        int nameStartOffset = pos;
        
        while (pos < source.length () && isXMLIdentifier (source.charAt (pos)))
            pos ++;
        
        if (pos == nameStartOffset)
            throw new XMLParseException ("Missing element name", token);
    }

    private void parseEndElement (Token token)
    {
        token.setType (Type.END_ELEMENT);
        
        skipWhiteSpace ();
        
        while (pos < source.length () && isXMLIdentifier (source.charAt (pos)))
            pos ++;
        
        skipWhiteSpace ();
        
        if (pos >= source.length () || source.charAt (pos ++) != '>')
            throw new XMLParseException ("Expected '>'", source, pos-1);
    }

    private void skipWhiteSpace ()
    {
        while (pos < source.length () && Character.isWhitespace (source.charAt (pos)))
            pos ++;
    }

    // <!--, <![CDATA[
    private void parseExcalamtion (Token token)
    {
        char c = source.charAt (pos);
        if (c == '-')
        {
            pos ++;
            parseComment (token);
        }
        else if (c == '[')
        {
            pos ++;
            parseCData (token);
        }
        else
            throw new XMLParseException ("Expected '<!--' or '<![CDATA['", source, pos-2);
    }

    private void parseCData (Token token)
    {
        token.setType (Type.CDATA);
        
        int errorPos = pos-3;
        if (pos >= source.length () || source.charAt (pos ++) != 'C')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);
        if (pos >= source.length () || source.charAt (pos ++) != 'D')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);
        if (pos >= source.length () || source.charAt (pos ++) != 'A')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);
        if (pos >= source.length () || source.charAt (pos ++) != 'T')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);
        if (pos >= source.length () || source.charAt (pos ++) != 'A')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);
        if (pos >= source.length () || source.charAt (pos ++) != '[')
            throw new XMLParseException ("Expected '<![CDATA['", source, errorPos);

        while (true)
        {
            if (pos >= source.length ())
                throw new XMLParseException ("Expected ']]>'", source, pos);
            
            char c = source.charAt (pos ++);
            if (c == ']')
            {
                if (pos >= source.length ())
                    throw new XMLParseException ("Expected ']]>'", source, pos-1);
                
                c = source.charAt (pos ++);
                if (c != ']')
                    continue;
                
                if (pos >= source.length ())
                    throw new XMLParseException ("Expected ']]>'", source, pos-2);
                
                c = source.charAt (pos ++);
                if (c == '>')
                    break;
            }
        }
    }

    private void parseComment (Token token)
    {
        token.setType (Type.COMMENT);
        
        if (pos >= source.length () || source.charAt (pos) != '-')
            throw new XMLParseException ("Expected '<!--'", source, pos-3);
        
        pos ++;
        
        while (true)
        {
            if (pos >= source.length ())
                throw new XMLParseException ("Expected '-->'", source, pos);
            
            char c = source.charAt (pos ++);
            if (c == '-')
            {
                if (pos >= source.length ())
                    throw new XMLParseException ("Expected '-->'", source, pos-1);
                
                c = source.charAt (pos ++);
                if (c != '-')
                    continue;
                
                if (pos >= source.length ())
                    throw new XMLParseException ("Expected '-->'", source, pos-2);
                
                c = source.charAt (pos ++);
                if (c != '>')
                    throw new XMLParseException ("XML comments can't contain '--'", source, pos-3);

                break;
            }
        }
    }

    private void parseProcessingInstruction (Token token)
    {
        token.setType (Type.PROCESSING_INSTRUCTION);
        
        while (true)
        {
            if (pos >= source.length ())
                throw new XMLParseException ("Expected '?>'", source, pos);
            
            char c = source.charAt (pos ++);
            if (c == '?')
            {
                if (pos >= source.length ())
                    throw new XMLParseException ("Expected '>'", source, pos);
                
                if (source.charAt (pos) == '>')
                {
                    pos ++;
                    break;
                }
            }
        }
    }

    private void parseAttribute (Token token)
    {
        token.setType (Type.ATTRIBUTE);

        while (pos < source.length () && isXMLIdentifier (source.charAt (pos)))
            pos ++;
        
        if (pos == token.getStartOffset())
            throw new XMLParseException ("Expected attribute name", source, pos);
        
        if (pos >= source.length () || source.charAt (pos) != '=')
            throw new XMLParseException ("Expected '='", source, pos);
        pos ++;
        
        char c = 0;
        if (pos < source.length ())
            c = source.charAt (pos);
        if (c != '\'' && c != '"')
            throw new XMLParseException ("Expected single or double quotes", source, pos);
        
        char endChar = c;
        
        while (true)
        {
            pos ++;
            if (pos >= source.length ())
                throw new XMLParseException ("Missing end quote ("+endChar+")", token);

            if (source.charAt (pos) == endChar)
                break;
        }
        
        // Skip end-char
        pos ++;
    }

    private boolean isXMLIdentifier (char c)
    {
        // Allowing ':' here is a cheap trick to make namespaces work
        return c == '-' 
            || c == '.' 
            || c == '_'
            || c == ':'
            || Character.isLetterOrDigit (c)
        ;
    }

    private void parseText (Token token)
    {
        token.setType (Type.TEXT);
        
        while (pos < source.length () && source.charAt (pos) != '<')
            pos ++;
    }
}
