package de.pdark.decentxml;

import java.io.ByteArrayInputStream;
import java.io.StringReader;
import java.util.List;
import java.util.Map;

import de.pdark.decentxml.XMLTokenizer.Type;

import junit.framework.TestCase;

public class XMLParserTest extends TestCase
{
    public void testRoundtrip () throws Exception
    {
        setUp (XMLTokenizerTest.XML);

        assertEquals (XMLTokenizerTest.XML, doc.toXML ());
    }
    
    public void testBuildDocument () throws Exception
    {
        Document doc = new Document ();
        doc.addNode (new ProcessingInstruction ("xml version=\"1.0\" encoding=\"utf-8\""));
        doc.addNode (new Text ("\n\n"));
        doc.addNode (new Comment (" Comment "));
        doc.addNode (new Text ("\n"));
        
        Element root = new Element ("root");
        doc.addNode (root);

        root.addChild (new Text ("\n    "));
        root.addChild (new Comment (" comment "));
        root.addChild (new Text ("\n    "));
        
        Element e = new Element ("e");
        root.addChild (e);
        e.addChild (new Text ("\n\ttext<>\n\t"));
        e.addChild (new Text (Type.CDATA, "text<></e>"));
        e.addChild (new Text ("\n    "));
        root.addChild (new Text ("\n    "));
        root.addChild (new Element ("empty"));
        root.addChild (new Text ("\n    "));
        e = new Element ("empty");
        root.addChild (e);
        e.addAttribute (new Text (Type.ELEMENT_WHITESPACE, " "));
        root.addChild (new Text ("\n    "));
        e = new Element ("a");
        root.addChild (e);
        
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("x", "1", '\''));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("y", "2"));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("f1", "\""));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("f2", "'"));
        e.addAttribute (new Text (" "));
        
        root.addChild (new Text ("\n    "));
        e = new Element ("a");
        root.addChild (e);

        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("x", "1", '\''));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("y", "2", '"'));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("f1", "\""));
        e.addAttribute (new Text (" "));
        e.addAttribute (new Attribute ("f2", "\"'"));
        e.addAttribute (new Text (" "));
        
        e.addChild (new Text ("b"));
        
        e.setBeginName (" a");
        e.setEndName (" a ");

        root.addChild (new Text ("\n"));
        doc.addNode (new Text ("\n"));
        
        assertEquals (XMLTokenizerTest.XML, doc.toXML ());
    }
    
    public void testNavigation () throws Exception
    {
        //System.out.println (XML);
        setUp (XMLTokenizerTest.XML);
        
        Element root = doc.getRootElement ();
        assertNotNull (root);
        
        Element e = root.getChild ("e");
        assertNotNull (e);
        assertEquals ("e", e.getName ());
    }
    
    public void testNavigation2 () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        Element root = doc.getRootElement ();
        List<Element> l = root.getChildren ("e");
        assertNotNull (l);
        assertEquals (1, l.size ());
        assertEquals (root.getChild ("e"), l.get (0));
    }
    
    public void testNavigation3 () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        Element root = doc.getRootElement ();
        List<Element> l = root.getChildren ("a");
        assertNotNull (l);
        assertEquals (2, l.size ());
    }
    
    public void testNavigation4 () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        Element root = doc.getRootElement ();
        List<Element> l = root.getChildren ("a");
        Element a = l.get (0);
        Attribute attr = a.getAttribute ("x");
        assertEquals ("x", attr.getName ());
        assertEquals ("1", attr.getAttributeValue ());
    }
    
    public void testDocumentType () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        assertEquals (Type.DOCUMENT, doc.getType ());
    }
    
    public void testElementType () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        assertEquals (Type.ELEMENT, doc.getRootElement ().getType ());
    }
    
    public void testElementChild () throws Exception
    {
        setUp ("<a/>");
        
        assertNull (doc.getRootElement ().getChild ("xxx"));
    }
    
    public void testCreateElement () throws Exception
    {
        Element e = new Element ("e");
        assertEquals (-1, e.getStartOffset ());
        assertEquals (-1, e.getEndOffset ());
    }
    
    public void testRemove () throws Exception
    {
        setUp (XMLTokenizerTest.XML);
        
        Node n;
        n = doc.remove (2);
        assertEquals (Type.COMMENT, n.getType ());
        n = doc.get (2);
        assertTrue (doc.remove (n));
        
        Element root = doc.getRootElement ();
        root.clearChildNodes ();
        
        root.addChild (new Text ("a"));
        root.addChild (new Text ("c"));
        root.addChild (1, new Text ("b"));
        
        assertEquals ("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" + 
                        "\n" + 
                        "<root>abc</root>\n" + 
                        "", doc.toXML ());
    }
    
    public void testAttributeMap () throws Exception
    {
        setUp ("<a/>");
        
        Map<String, Attribute> map = doc.getRootElement ().getAttributeMap ();
        assertEquals ("{}", map.toString ());
    }
    
    public void testAttributeMap2 () throws Exception
    {
        setUp ("<a />");
        
        Map<String, Attribute> map = doc.getRootElement ().getAttributeMap ();
        assertEquals ("{}", map.toString ());
    }
    
    public void testAttributeMap3 () throws Exception
    {
        setUp ("<a x='1' y='2' />");
        
        Map<String, Attribute> map = doc.getRootElement ().getAttributeMap ();
        assertEquals ("{x=x='1', y=y='2'}", map.toString ());
    }
    
    public void testAttributeMapSequence () throws Exception
    {
        setUp ("<a y='2' x='1' />");
        
        Map<String, Attribute> map = doc.getRootElement ().getAttributeMap ();
        assertEquals ("{y=y='2', x=x='1'}", map.toString ());
    }
    
    public void testAttributes () throws Exception
    {
        setUp ("<a/>");
        
        List<Attribute> list = doc.getRootElement ().getAttributes ();
        assertEquals ("[]", list.toString ());
    }
    
    public void testAttributes2 () throws Exception
    {
        setUp ("<a />");
        
        List<Attribute> list = doc.getRootElement ().getAttributes ();
        assertEquals ("[]", list.toString ());
    }
    
    public void testAttributes3 () throws Exception
    {
        setUp ("<a y='2' x='1' />");
        
        List<Attribute> list = doc.getRootElement ().getAttributes ();
        assertEquals ("[y='2', x='1']", list.toString ());
    }
    
    public void testGetAttribute () throws Exception
    {
        setUp ("<a/>");
        
        Attribute a = doc.getRootElement ().getAttribute ("x");
        assertNull (a);
    }
    
    public void testGetAttribute2 () throws Exception
    {
        setUp ("<a />");
        
        Attribute a = doc.getRootElement ().getAttribute ("x");
        assertNull (a);
    }
    
    public void testGetAttribute3 () throws Exception
    {
        setUp ("<a y='2' x='1' />");
        
        Attribute a = doc.getRootElement ().getAttribute ("x");
        assertEquals ("x='1'", a.toString ());
    }
    
    public void testGetAttributeValue () throws Exception
    {
        setUp ("<a/>");
        
        String a = doc.getRootElement ().getAttributeValue ("x");
        assertNull (a);
    }
    
    public void testGetAttributeValue2 () throws Exception
    {
        setUp ("<a />");
        
        String a = doc.getRootElement ().getAttributeValue ("x");
        assertNull (a);
    }
    
    public void testGetAttributeValue3 () throws Exception
    {
        setUp ("<a y='2' x='1' />");
        
        String a = doc.getRootElement ().getAttributeValue ("x");
        assertEquals ("1", a.toString ());
    }
    
    public void testAddChild () throws Exception
    {
        Element e = new Element ("a");
        Attribute a = new Attribute ("x", "1");
        try
        {
            e.addChild (a);
            fail ("No exception was thrown");
        }
        catch (XMLParseException ex)
        {
            assertEquals ("Line 1, column 1: Can't add an attribute as a child of an element", ex.getMessage ());
        }
    }
    
    public void testIsCompactEmpty () throws Exception
    {
        setUp ("<a/>");
        assertTrue (doc.getRootElement ().isCompactEmpty ());
    }
    
    public void testIsCompactEmpty2 () throws Exception
    {
        setUp ("<a />");
        assertTrue (doc.getRootElement ().isCompactEmpty ());
    }
    
    public void testIsCompactEmpty3 () throws Exception
    {
        setUp ("<a></a>");
        assertFalse (doc.getRootElement ().isCompactEmpty ());
        assertEquals ("<a></a>", doc.toXML ());
    }
    
    public void testIsCompactEmpty4 () throws Exception
    {
        setUp ("<a />");
        doc.getRootElement ().addChild (new Element ("e"));
        assertFalse (doc.getRootElement ().isCompactEmpty ());
    }
    
    public void testGetChildNodes () throws Exception
    {
        setUp ("<a />");
        assertEquals ("[]", doc.getRootElement ().getNodes ().toString ());
    }
    
    public void testGetChildNodes2 () throws Exception
    {
        setUp ("<a> </a>");
        assertEquals ("[ ]", doc.getRootElement ().getNodes ().toString ());
    }
    
    public void testRemoveChildNode () throws Exception
    {
        setUp ("<a> <b/></a>");
        Node n = doc.getRootElement ().removeNode (0);
        assertNotNull (n);
        assertEquals ("<a><b/></a>", doc.toXML ());
    }

    public void testRemoveChildNode2 () throws Exception
    {
        setUp ("<a> <b/></a>");
        Node n = doc.getRootElement ().removeNode (1);
        assertNotNull (n);
        assertEquals ("<a> </a>", doc.toXML ());
    }
    
    public void testRemoveChildNode3 () throws Exception
    {
        setUp ("<a> <b/></a>");
        Node n = doc.getRootElement ().getNode (0);
        assertTrue (doc.getRootElement ().removeNode (n));
        assertEquals ("<a><b/></a>", doc.toXML ());
    }
    
    public void testRemoveChildNode4 () throws Exception
    {
        Element e = new Element ("e");
        assertFalse (e.removeNode (null));
    }
    
    public void testGetChildren () throws Exception
    {
        setUp ("<a />");
        assertEquals ("[]", doc.getRootElement ().getChildren ().toString ());
    }
    
    public void testGetChildren2 () throws Exception
    {
        setUp ("<a> <b/></a>");
        assertEquals ("[<b/>]", doc.getRootElement ().getChildren ().toString ());
    }
    
    public void testGetText () throws Exception
    {
        setUp ("<a />");
        assertEquals ("", doc.getRootElement ().getText ());
    }
    
    public void testGetText2 () throws Exception
    {
        setUp ("<a>  x \n y  </a>");
        assertEquals ("  x \n y  ", doc.getRootElement ().getText ());
    }
    
    public void testGetText3 () throws Exception
    {
        setUp ("<a>a<b>x</b>a</a>");
        assertEquals ("axa", doc.getRootElement ().getText ());
    }
    
    public void testGetTrimmedText () throws Exception
    {
        setUp ("<a>  x \n y  </a>");
        assertEquals ("x \n y", doc.getRootElement ().getTrimmedText ());
    }
    
    public void testGetNormalizedText () throws Exception
    {
        setUp ("<a> <b> x \n y </b> </a>");
        assertEquals ("x y", doc.getRootElement ().getNormalizedText ());
    }
    
    public void testClearText () throws Exception
    {
        setUp ("<a />");
        doc.getRootElement ().clearText ();
        assertEquals ("<a />", doc.toXML ());
    }
    
    public void testClearText2 () throws Exception
    {
        setUp ("<a> </a>");
        doc.getRootElement ().clearText ();
        assertEquals ("<a></a>", doc.toXML ());
    }
    
    public void testClearText3 () throws Exception
    {
        setUp ("<a> <b>x</b> </a>");
        doc.getRootElement ().clearText ();
        assertEquals ("<a><b></b></a>", doc.toXML ());
    }
    
    public void testClearText4 () throws Exception
    {
        setUp ("<a> <b>x</b> </a>");
        doc.getRootElement ().getChild ("b").clearText ();
        assertEquals ("<a> <b></b> </a>", doc.toXML ());
    }
    
    String POM_XML = 
        "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n" + 
        "\r\n" + 
        "<project xmlns=\"http://maven.apache.org/POM/4.0.0\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://maven.apache.org/POM/4.0.0 http://maven.apache.org/maven-v4_0_0.xsd\">\r\n" + 
        "  <modelVersion>4.0.0</modelVersion>\r\n" + 
        "\r\n" + 
        "  <!-- Ignore this <parent>\r\n" + 
        "    <groupId>org.codehaus.mojo</groupId>\r\n" + 
        "    <artifactId>mojo</artifactId>\r\n" + 
        "    <version>15</version>\r\n" + 
        "  </parent> -->\r\n" + 
        "  <parent>\r\n" + 
        "    <groupId>org.codehaus.mojo</groupId>\r\n" + 
        "    <artifactId>mojo</artifactId>\r\n" + 
        "    <version>16</version>\r\n" + 
        "  </parent>\r\n" + 
        "\r\n" + 
        "  <groupId>org.codehaus.mojo</groupId>\r\n" + 
        "  <artifactId>versions-maven-plugin</artifactId>\r\n" + 
        "  <version>1.0-SNAPSHOT</version>\r\n" + 
        "  <packaging>maven-plugin</packaging>\r\n" + 
        "</project>\r\n" + 
        "";
        
    public void testNothingFound () throws Exception
    {
        setUp (POM_XML);

        Element match = root.getChild ("xxx");
        assertNull (match == null ? "" : match.toString (), match);
    }
    
    public void testProject () throws Exception
    {
        setUp (POM_XML);

        assertEquals ("project", root.getName ());
        assertEquals (42, root.getStartToken ().getStartOffset());
    }
    
    public void testParent () throws Exception
    {
        setUp (POM_XML);

        Element match = root.getChild ("parent");
        assertEquals (437, match.getStartOffset());
        assertEquals (562, match.getEndOffset());
        
        assertEquals (
                "<parent>\r\n" + 
                "    <groupId>org.codehaus.mojo</groupId>\r\n" + 
                "    <artifactId>mojo</artifactId>\r\n" + 
                "    <version>16</version>\r\n" + 
                "  </parent>", 
                match.getStartToken ().getSource ().substring (match.getStartOffset (), match.getEndOffset ()));
    }
    
    public void testParentVersion () throws Exception
    {
        setUp (POM_XML);

        Element match = root.getChild ("parent");
        Element version = match.getChild ("version");
        assertEquals (528, version.getStartToken ().getStartOffset());
        assertEquals ("<version>16</version>", version.toXML ());
        assertEquals ("16", version.getNormalizedText ());
    }
    
    public void testReplaceParentVersion () throws Exception
    {
        setUp (POM_XML);

        Element match = root.getChild ("parent");
        Element version = match.getChild ("version");
        version.setText ("17");
        
        assertEquals (POM_XML.replaceAll ("16", "17"), doc.toXML ());
    }
    
    public void testParentXXX () throws Exception
    {
        setUp (POM_XML);

        Element parent = root.getChild ("parent");
        Element match = parent.getChild ("xxx");
        assertNull (match == null ? "" : match.toString (), match);
    }
    
    public void testProjectArtifactId () throws Exception
    {
        setUp (POM_XML);

        Element artifactId = root.getChild ("artifactId");
        assertEquals ("versions-maven-plugin", artifactId.getNormalizedText ());
    }
    
    public void testNoAttributeXXX () throws Exception
    {
        setUp ("<a />");
        Attribute a = root.getAttribute ("xxx");
        assertNull (a);
    }
    
    public void testNoAttributeXXX2 () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("xxx");
        assertNull (a);
    }
    
    public void testSetAttributeName () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        a.setName ("y");
        assertEquals ("<a y='1' />", doc.toXML ());
    }
    
    public void testSetAttributeValue () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        a.setAttributeValue ("2");
        assertEquals ("<a x='2' />", doc.toXML ());
    }
    
    public void testSetAttributeValue2 () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        a.setAttributeValue ("\"x\"");
        assertEquals ("<a x='\"x\"' />", doc.toXML ());
    }
    
    public void testSetAttributeValue3 () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        a.setAttributeValue ("'b'");
        assertEquals ("<a x=\"'b'\" />", doc.toXML ());
    }
    
    public void testSetAttributeValue4 () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        a.setAttributeValue ("'\"");
        assertEquals ("<a x=\"'&quot;\" />", doc.toXML ());
    }
    
    public void testAttributeToXML () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        assertEquals ("x='1'", a.toXML ());
    }
    
    public void testAttributeType () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        assertEquals (Type.ATTRIBUTE, a.getType ());
    }
    
    public void testAttributeOffsets () throws Exception
    {
        setUp ("<a x='1' />");
        Attribute a = root.getAttribute ("x");
        assertEquals (3, a.getStartOffset ());
        assertEquals (8, a.getEndOffset ());
    }
    
    public void testNoRootElement () throws Exception
    {
        try
        {
            setUp ("<?xml?>");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 1, column 1: No root element found", e.getMessage ());
        }
    }
    
    public void testTwoRootElements () throws Exception
    {
        try
        {
            setUp ("<?xml?><a /><b/>");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 1, column 13: Only one root element allowed per document", e.getMessage ());
        }
    }
    
    public void testUnexpectedEOF () throws Exception
    {
        try
        {
            setUp ("<?xml?>\n<a");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 2, column 3: Unexpected end-of-file while parsing attributes of element a", e.getMessage ());
        }
    }
    
    public void testUnexpectedEOF2 () throws Exception
    {
        try
        {
            setUp ("<?xml?>\n<a>");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 2, column 4: Unexpected end-of-file while parsing children of element a", e.getMessage ());
        }
    }
    
    public void testElementMismatch () throws Exception
    {
        try
        {
            setUp ("<?xml?>\n<a></b>");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 2, column 1: End element 'b' doesn't match with 'a'", e.getMessage ());
        }
    }
    
    public void testElementMismatch2 () throws Exception
    {
        try
        {
            setUp ("<?xml?>\n<a>\n<b></b>\n<b>\n<b></b>\n</a>");
            fail ("No exception was thrown");
        }
        catch (XMLParseException e)
        {
            assertEquals ("Line 4, column 1: End element 'a' doesn't match with 'b'", e.getMessage ());
        }
    }
    
    public void testXMLParseException () throws Exception
    {
        XMLParseException e = new XMLParseException ("x", (XMLTokenizer.Token)null);
        assertEquals ("Line 1, column 1: x", e.getMessage ());
        assertNull (e.getSource ());
    }
    
    public void testXMLParseException2 () throws Exception
    {
        XMLSource source = new XMLStringSource ("a\rb\r\nc\nd");
        XMLParseException e = new XMLParseException ("x", source, source.length () - 1);
        assertEquals (7, e.getOffset ());
        assertEquals ("Line 4, column 1: x", e.getMessage ());
        assertEquals (source, e.getSource ());
    }
    
    public void testXMLParseException3 () throws Exception
    {
        XMLSource source = new XMLStringSource ("a\rb\r\nc\nd");
        XMLParseException e = new XMLParseException ("x", source, source.length ());
        assertEquals (2, e.getColumn ());
        assertEquals ("Line 4, column 2: x", e.getMessage ());
    }
    
    public void testXMLParseException4 () throws Exception
    {
        XMLSource source = new XMLStringSource ("a\rb\r\nc\n");
        XMLParseException e = new XMLParseException ("x", source, source.length ());
        assertEquals (7, e.getLineStartOffset ());
        assertEquals ("Line 4, column 1: x", e.getMessage ());
    }
    
    public void testXMLParseException5 () throws Exception
    {
        XMLSource source = new XMLStringSource ("a\rb\r\nc\r\n");
        XMLParseException e = new XMLParseException ("x", source, source.length ());
        assertEquals (8, e.getLineStartOffset ());
        assertEquals ("Line 4, column 1: x", e.getMessage ());
    }
    
    public void testXMLParseException6 () throws Exception
    {
        XMLSource source = new XMLStringSource ("a\rb\r\nc\r");
        XMLParseException e = new XMLParseException ("x", source, source.length ());
        assertEquals (7, e.getLineStartOffset ());
        assertEquals ("Line 4, column 1: x", e.getMessage ());
    }

    public void testXMLStringBufferSource () throws Exception
    {
        StringBuffer buffer = new StringBuffer ("<x a='1'>n</x>");
        XMLParser parser = new XMLParser ();
        doc = parser.parse (new XMLStringBufferSource (buffer));
        assertEquals (buffer.toString (), doc.toXML ());
    }

    public void testXMLStringBuilderSource () throws Exception
    {
        StringBuilder buffer = new StringBuilder ("<x a='1'>n</x>");
        XMLParser parser = new XMLParser ();
        doc = parser.parse (new XMLStringBuilderSource (buffer));
        assertEquals (buffer.toString (), doc.toXML ());
    }
    
    public void testXMLIOSource1 () throws Exception
    {
        XMLIOSource source = new XMLIOSource (new ByteArrayInputStream (XMLInputStreamReaderTest.TEST_XML.getBytes ("utf-8")));
        XMLParser parser = new XMLParser ();
        doc = parser.parse (source);
        assertEquals (XMLInputStreamReaderTest.TEST_XML, doc.toXML ());
    }

    public void testXMLIOSource2 () throws Exception
    {
        XMLIOSource source = new XMLIOSource (new StringReader (XMLInputStreamReaderTest.TEST_XML));
        XMLParser parser = new XMLParser ();
        doc = parser.parse (source);
        assertEquals (XMLInputStreamReaderTest.TEST_XML, doc.toXML ());
    }
    
    private Document doc;
    private Element root;

    public void setUp (String xml)
    {
        XMLParser parser = new XMLParser ();
        doc = parser.parse (new XMLStringSource (xml));
        root = doc.getRootElement ();
    }
    
    @Override
    protected void setUp () throws Exception
    {
        super.setUp ();
    }
    
    @Override
    protected void tearDown () throws Exception
    {
        doc = null;
        root = null;
        
        super.tearDown ();
    }
}
